#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
sys.path.insert(0, os.path.abspath('../'))
import pykafka_tools


__author__ = u'Stephan Müller'
__copyright__ = u'2017, Stephan Müller'
__license__ = u'MIT'

version = release = pykafka_tools.__version__

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.coverage',
]

templates_path = ['_templates']
exclude_patterns = ['_build']
html_static_path = ['_static']

source_suffix = '.rst'
master_doc = 'index'

html_theme = 'sphinx_rtd_theme'
pygments_style = 'sphinx'
htmlhelp_basename = 'pykafka-tools'

autodoc_default_flags = ['special-members', 'private-members', 'show-inheritance']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import io
import avro
import avro.io

__author__ = u'Stephan Müller'
__copyright__ = u'2017, Stephan Müller'
__license__ = u'MIT'


class AvroCoding(object):

    def __init__(self, kafka_message_schema_file):
        self._kafka_message_schema = avro.schema.Parse(open(kafka_message_schema_file, "rb").read().decode())

    def decode(self, message):
        bytes_reader = io.BytesIO(message)
        decoder = avro.io.BinaryDecoder(bytes_reader)
        reader = avro.io.DatumReader(self._kafka_message_schema)
        return reader.read(decoder)

    def encode(self, message):
        bytes_writer = io.BytesIO()
        encoder = avro.io.BinaryEncoder(bytes_writer)
        writer = avro.io.DatumWriter(self._kafka_message_schema)
        writer.write(message, encoder)
        return bytes_writer.getvalue()



# -*- coding: utf-8 -*-

import logging
import os
from threading import Thread
import avro.io
from pykafka import KafkaClient
import pykafka.exceptions
from pykafka_tools.avro_coding import AvroCoding

__author__ = u'Stephan Müller'
__copyright__ = u'2017, Stephan Müller'
__license__ = u'MIT'

logger = logging.getLogger(__name__)


class KafkaProducerPool(object):
    """
    Manages connection to Kafka and is responsible for producing messages. Holds a cache of messages if connection
    to Kafka is lost and send them after connection is established again.
    """

    def __init__(self, kafka_hosts, topics, kafka_message_schema_file=None, use_rdkafka=False, sync=True):
        """
        Instantiate a new KafkaProducerPool and tries to connect to Kafka

        :param kafka_hosts: kafka hosts
        :type kafka_hosts: str
        :param topics: list of topics for which a producer should being created
        :type topics: list of str
        :param use_rdkafka: set true to use rdkafka
        :type use_rdkafka: bool
        :param sync: If true, directly sends messages to Kafka
        :type sync: bool
        """
        self._kafka_hosts = kafka_hosts
        self._topic_names = topics
        self._use_rdkafka = use_rdkafka
        self._sync = sync
        self._avro_coding = None
        if kafka_message_schema_file is not None:
            if os.path.isfile(kafka_message_schema_file):
                self._avro_coding = AvroCoding(kafka_message_schema_file)

        self._client = None
        self._topics = dict()
        self._producers = dict()
        self._connected = False
        self._try_reconnect = False
        self._producer_is_working = False
        self._producer_messages = dict()
        for topic in topics:
            self._producer_messages.update({
                topic: list()
            })
        self.connect()

    def _reset(self):
        """
        Delete all Kafka clients, topics and producers. Call this method after connection was lost.
        """
        self._client = None
        self._topics = dict()
        self._producers = dict()

    def connect(self):
        """
        Tries to connect to Kafka by creating a new thread so that main thread is not being blocked.
        On successful connection, producers for selected topics are created.
        """
        if self._try_reconnect is False:
            logger.debug("Try connecting.")

            def connect_thread():
                self._try_reconnect = True
                try:
                    self._client = KafkaClient(hosts=self._kafka_hosts)
                    for topic_name in self._topic_names:
                        self._topics.update({
                            topic_name: self._client.topics[topic_name.encode()]
                        })
                        self._producers.update({
                            topic_name:
                                self._topics[topic_name].get_producer(use_rdkafka=self._use_rdkafka, sync=self._sync)
                        })
                    self._connected = True
                except Exception as e:
                    self._reset()
                    logger.exception(e)
                self._try_reconnect = False

            thread = Thread(target=connect_thread)
            thread.start()
        else:
            logger.debug("Already tries to connect.")

    def produce(self, topic, message):
        """
        Collects messages to produce and triggers a new thread to actually send messages so that main thread will
        not being blocked.

        :param topic: topic name
        :type topic: str
        :param message: message
        :type message: str or bytes or dict
        """
        try:
            if type(message) is bytes:
                pass
            if type(message) is dict and self._avro_coding is not None:
                try:
                    message = self._avro_coding.encode(message)

                except (AssertionError, IndexError, avro.io.SchemaResolutionException):
                    logger.warning("Could not parse message of topic '" + topic + "'. Continuing.")
                    return
            else:
                message = str(message).encode()

            self._producer_messages[topic].append(message)

        except KeyError:
            raise KeyError("Producer for given topic '" + topic + "' is not initialized. Please make sure that all"
                           " desired topics are listed in KafkaProducerPool.__init__()")
        if self._connected is False and self._try_reconnect is True:
            logger.debug("Not connected to Kafka but still trying to connect.")
        elif self._connected is False:
            self.connect()
        else:
            def produce_thread():
                self._producer_is_working = True
                sent_messages_count = 0
                processed_messages_count = 0
                try:
                    for topic_name in self._producer_messages:
                        while len(self._producer_messages[topic_name]) > 0:
                            processed_messages_count += 1
                            self._producers[topic_name].produce(self._producer_messages[topic_name][0])
                            self._producer_messages[topic_name].pop(0)
                            sent_messages_count += 1
                except pykafka.exceptions.NoBrokersAvailableError:
                    self._connected = False
                    logger.error("Disconnected because connection to brokers could not be established.")
                except pykafka.exceptions.RdKafkaException as e:
                    self._connected = False
                    logger.error(e)
                except Exception as e:
                    logger.exception(e)
                    raise
                finally:
                    self._producer_is_working = False
                    logger.info("Sent %s of %s messages successful to Kafka.", sent_messages_count,
                                processed_messages_count)

            if not self._producer_is_working:
                thread = Thread(name='Producer', target=produce_thread)
                thread.start()
            else:
                logger.debug("Another produce thread is still working.")



# -*- coding: utf-8 -*-

import logging
import os
import threading
import datetime
import time
import avro.io
import psycopg2
import pykafka.exceptions
from pykafka import KafkaClient
import pykafka_tools.event as event
from pykafka_tools.avro_coding import AvroCoding

__author__ = u'Stephan Müller'
__copyright__ = u'2017, Stephan Müller'
__license__ = u'MIT'

logger = logging.getLogger(__name__)


class KafkaConsumer(threading.Thread):
    """A KafkaConsumer is a thread which consumes messages and notifies functions by an event trigger after every new
     consumption.
     """

    def __init__(self, kafka_hosts, topic, consumer_group, kafka_message_schema_file=None, use_rdkafka=False,
                 auto_commit_enable=True, auto_commit_interval_ms=60000):
        """
        Initializes a new instance

        :param kafka_hosts: kafka hosts
        :type kafka_hosts: str
        :param topic: topic for which a consumer should being created
        :type topic: str
        :param consumer_group: consumer group
        :type consumer_group: str
        :param use_rdkafka: set true to use rdkafka
        :type use_rdkafka: bool
        :param auto_commit_enable: enable auto commit
        :type auto_commit_enable: bool
        :param auto_commit_interval_ms: auto commit interval in milliseconds
        :type auto_commit_interval_ms: int
        :param kafka_message_schema_file: kafka message schema. If messages are avro encoded, the schema is used to
               decode the messages.
        :type kafka_message_schema_file: file str
        """

        super().__init__(name='KafkaConsumer_' + topic)
        self._stopevent = threading.Event()
        self._topic = topic
        self._consumer_group = consumer_group
        self._kafka_hosts = kafka_hosts
        self._use_rdkafka = use_rdkafka
        self._auto_commit_enable = auto_commit_enable
        self._auto_commit_interval_ms = auto_commit_interval_ms
        self._avro_coding = None
        if kafka_message_schema_file is not None:
            if os.path.isfile(kafka_message_schema_file):
                self._avro_coding = AvroCoding(kafka_message_schema_file)

        self._consumer = None

        self.new_message_event = event.Event()

    def get_topic(self):
        """
        Get topic name
        :return: topic name
        :type: str
        """
        return self._topic

    def run(self):
        """ Thread loop. If establishing connection fails, a new connection try is performed every second.
        """

        while not self._stopevent.isSet():
            try:
                client = KafkaClient(hosts=self._kafka_hosts)
                topic = client.topics[str.encode(self._topic)]
                self._consumer = topic.get_simple_consumer(consumer_group=str.encode(self._consumer_group),
                                                           auto_commit_enable=self._auto_commit_enable,
                                                           auto_commit_interval_ms=self._auto_commit_interval_ms,
                                                           use_rdkafka=self._use_rdkafka)
                for message in self._consumer:
                    try:
                        if self._stopevent.isSet():
                            raise InterruptedError

                        if self._avro_coding is None:
                            message_data = message.value

                        else:
                            try:
                                message_data = self._avro_coding.decode(message.value)

                            except (AssertionError, IndexError, avro.io.SchemaResolutionException):
                                logger.warning("Could not parse message with offset " + str(message.offset) +
                                               " of topic '" + self._topic + "'. Continuing.")
                                continue

                        log_message = "Received message from topic " + self._topic + " with offset "\
                                      + str(message.offset)
                        if type(message_data) == dict and "timestamp" in message_data:
                            log_message += " and timestamp " + datetime.datetime.fromtimestamp(
                                message_data["timestamp"]).strftime('%Y-%m-%d %H:%M:%S.%f')

                        logger.info(log_message)

                        self.new_message_event.fire(self, message_data)

                    except pykafka.exceptions.ConsumerStoppedException:
                        logger.warning("Consumer for topic " + self._topic + " was stopped manually.")

                    except Exception as e:
                        logger.exception(e)

            except pykafka.exceptions.ConsumerStoppedException:
                logger.warning("Consumer for topic " + self._topic + " was stopped manually.")

            except Exception as e:
                logger.exception(e)
                time.sleep(1)

    def stop(self):
        self._stopevent.set()
        if self._consumer is not None:
            self._consumer.stop()

    def join(self, timeout=None):
        """
        Stop the thread loop and wait for it to end.

        :param timeout: timeout
        :type timeout: float
        """
        self.stop()
        super().join(timeout)


class DBWriter(KafkaConsumer):

    def __init__(self, kafka_hosts, topic, consumer_group, kafka_message_schema_file, postgres_connector,
                 use_rdkafka=False, auto_commit_enable=True, auto_commit_interval_ms=60000):

        super().__init__(kafka_hosts, topic, consumer_group, kafka_message_schema_file, use_rdkafka,
                         auto_commit_enable, auto_commit_interval_ms)

        if self._avro_coding is None:
            raise IOError("Kafka message schema file does not exist.")

        self._postgres_connector = postgres_connector
        self.new_message_event += self._new_message_handler
        self._parsed_first_message_correctly = False

    def _new_message_handler(self, sender, message):

        try:
            if not self._parsed_first_message_correctly:
                # create or update table schema if necessary
                if not self._postgres_connector.table_exists(self._topic):
                    self._postgres_connector.create_table(table_name=self._topic, data=message["data"])
                    logger.info("Created table for topic '" + self._topic + "'")

                else:
                    number_of_added_columns = self._postgres_connector.update_columns(table_name=self._topic,
                                                                                      data=message["data"])
                    logger.info("Added " + str(number_of_added_columns) + " columns in table for topic '" + self._topic + "'")
                self._parsed_first_message_correctly = True

            self._postgres_connector.insert_values(table_name=self._topic, data=message)

        except psycopg2.OperationalError:
            raise ConnectionError("Postgres operational error.")

        except:
            # TODO not a nice way, because another exception is thrown
            self.join(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging

import psycopg2
import psycopg2.extras

__author__ = u'Stephan Müller'
__copyright__ = u'2017, Stephan Müller'
__license__ = u'MIT'


logger = logging.getLogger(__name__)


_data_types = {
    bytes: "bytea",
    float: "real",
    int: "bigint",
    bool: "bool",
    str: "text"
}


class PostgresConnector:

    def __init__(self, host, port, database, user, password):
        self._host = host
        self._port = port
        self._database = database
        self._user = user
        self._password = password

        try:
            self.con = psycopg2.connect(host=host, port=port, database=database, user=user, password=password)
            self.con.autocommit = True
        except Exception as e:
            raise e

    def table_exists(self, table_name):
        cur = self.con.cursor()
        sql = cur.mogrify("""
                          SELECT 1
                          FROM information_schema.tables
                          WHERE table_name=%s;
                          """, (table_name,))
        cur.execute(sql)

        result = cur.fetchall()
        if len(result) == 1:
            return True
        return False

    def create_table(self, table_name, data):

        cur = self.con.cursor()

        column_names = list()
        column_sql = ""
        for key in data:
            column_names.append(key)

            try:
                data_type = _data_types[type(data[key])]
            except KeyError:
                data_type = _data_types[str]

            column_sql += ", " + key + " " + data_type

        sql = """
              CREATE TABLE IF NOT EXISTS %s (
                timestamp TIMESTAMP WITH TIME ZONE PRIMARY KEY DEFAULT current_timestamp
                %s
              );
              """ % (table_name, column_sql)

        try:
            cur.execute(sql)
            self.con.commit()

        except psycopg2.DatabaseError as de:
            logger.error(de.pgcode + " " + de.pgerror.replace("\n", " "))

    def update_columns(self, table_name, data):

        cur = self.con.cursor()
        sql = cur.mogrify("""
                          SELECT column_name
                          FROM information_schema.columns
                          WHERE table_name=%s;
                          """, [table_name])
        cur.execute(sql)

        table_columns = cur.fetchall()[1:]

        column_names_to_add = list()
        for key in data:
            # if key does not exist
            if len([x[0] for x in table_columns if x[0] == key.lower()]) == 0:
                column_names_to_add.append(key.lower())

        for column_name in column_names_to_add:

            try:
                data_type = _data_types[type(data[column_name])]
            except KeyError:
                data_type = _data_types[str]

            sql = "ALTER TABLE %s ADD COLUMN %s %s;" % (table_name, column_name, data_type)

            try:
                cur.execute(sql)

            except psycopg2.DatabaseError as e:
                logger.error(e.pgcode + " " + e.pgerror.replace("\n", " "))

        self.con.commit()

        return len(column_names_to_add)

    def insert_values(self, table_name, data):

        cur = self.con.cursor()
        # empty string list item that there will be a comma after column timestamp
        columns = [""]
        values = list()
        for key in data["data"]:
            columns.append(key)
            values.append(data["data"][key])

        params = [data["timestamp"]]
        params.extend(values)
        sql = cur.mogrify("INSERT INTO " + table_name + " (timestamp" + ', '.join(columns) + ")"
                          " VALUES (to_timestamp(%s)" + ', %s' * len(values) + ");", params)

        try:
            cur.execute(sql)
            self.con.commit()

        except psycopg2.OperationalError as e:
            logger.error(e)
            raise e

        except psycopg2.DatabaseError as e:

            if e.pgcode == "23505":
                logger.warning(e.pgcode + " " + e.pgerror.replace("\n", " "))
                self.con.commit()
            else:
                logger.error(e.pgcode + " " + e.pgerror.replace("\n", " "))



# -*- coding: utf-8 -*-

__author__ = u'Stephan Müller'
__copyright__ = u'2016, Stephan Müller'
__license__ = u'MIT'


class Event(object):
    def __init__(self):
        self.handlers = []

    def add(self, handler):
        self.handlers.append(handler)
        return self

    def remove(self, handler):
        self.handlers.remove(handler)
        return self

    def fire(self, sender, *args):
        for handler in self.handlers:
            handler(sender, *args)

    __iadd__ = add
    __isub__ = remove



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup
import pykafka_tools

__author__ = u'Stephan Müller'
__copyright__ = u'2016, Stephan Müller'
__license__ = u'MIT'


setup(
    name='pykafka_tools',
    packages=['pykafka_tools'],
    version=pykafka_tools.__version__,
    license='MIT',
    description='The pykafka-tools extend the functionalities of python pykafka module.',
    author='Stephan Müller',
    author_email='stephan@mueller5.eu',
    url='https://github.com/smueller18/pykafka-tools',
    download_url='https://github.com/smueller18/pykafka-tools',
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved :: MIT License"
    ],
    install_requires=[
        'pykafka==2.6.0.dev3'
    ],
    dependency_links=[
        'https://github.com/smueller18/pykafka/tarball/2.6.0.dev3#egg=pykafka-2.6.0.dev3'
    ],
)
